; ide.asm - IDE hard disk driver

; Documentation:
;	/comp/doc/disk/ide/ide_tech.txt

; To do:
;	Issue SEEK command when needed (ie, multisector writes)
;	Do more error checking, reliability stuff, etc.

; Notes:
;	Sector addressing: Use LBA if available, CHS otherwise.


; Ports:
;; Command block registers
cbr
.data		equ	0x1f0	;r/w	(16 bits)
.error		equ	0x1f1	;r
.precomp	equ	0x1f1	;  w	or features (ATA-2)
.sec_count	equ	0x1f2	;r/w
.sec_num	equ	0x1f3	;r/w			LBA: bits 0-7 (LSB)
.cyl_l		equ	0x1f4	;r/w			LBA: bits 8-15
.cyl_h		equ	0x1f5	;r/w			LBA: bits 16-23
.drive_head	equ	0x1f6	;r/w	1L1Dhhhh	LBA: bits 24-27 (instead of head)
				;	 ^ L: 0=CHS, 1=LBA   D: 0=master, 1=slave
.status		equ	0x1f7	;r
.command	equ	0x1f7	;  w

;; Control block registers
%define alt_status	0x3f6	;r
%define control		0x3f6	;  w
%define drv_addr	0x3f7	;r

	SECTION .data
regs
.pc	db	1	; write precompensation
.sc	db	1	; sector count
.lba
.sn	db	1	; sector number
.cy	dw	0	; cylinder number
.dh	db	0xA0	; drive/head

valid	db	0	; 1 = dcyl, dhead, dsect (below) are valid

	SECTION .bss
dcyl	resd	1	; # of cylinders
dhead	resd	1	; # of heads
dsect	resd	1	; # of sectors
done	resb	1	; Done (IRQ received) flag

	SECTION .text
init_ide:
	sprint	"IDE hard disk driver",10
	callproc patch_int, 0x20+14, irq14_handler
	in	al, 0xA1				; enable IRQ 14
	and	al, 10111111b
	out	0xA1, al
	call	ide_id		; Read disk specs
	ret

irq14_handler:
;	pchr '%'
	inc byte [done]
	push	eax
	irqack	14
	pop	eax
	sti
	iretd


;;
;; TODO: Timeouts for all these waits
;;

ide_wait_irq:		; Wait for IRQ14
.1	cmp byte [done], 0
	jz	.1
	ret


ide_busy_wait:		; Wait for -BUSY
	mov	dx, cbr.status
.1	in	al, dx
	test	al, 0x80
	jnz	.1
	ret


ide_drdy_wait:		; Wait for DRDY
	mov	dx, cbr.status
.1	in	al, dx
	test	al, 0x40
	jz	.1
	ret


ide_wait_drq:		; Wait for DRQ (Data ReQuest)
	mov	dx, cbr.status
.1	in	al, dx
	test	al, 0x08
	jz	.1
	ret


ide_cmd:	;; ( cmd-byte -- )
;;
;; Procedure:
;; 1. Wait for -BUSY - bit 7 (0x80) of status reg. 1f7 (wait for it to go OFF)
;; 2. Load command block regs.
;; 3. Activate -IEN (interrupt enable) - turn OFF bit 1 (0x02) of control reg. 3f6
;;	(IRQs 2 (cascade) and 14 (ide) must be unmasked in regs 0x21 and 0xA1)
;; 4. Wait for DRDY - bit 6 (0x40) of status reg. 1f7
;; 5. Write command code to command reg. 1f7
;; After that it depends upon the command...
;;
	mpush	eax, ecx, edx, esi
	call	ide_busy_wait
	mov	dx, cbr.precomp	; load command block registers (from regs.*)
	mov	esi, regs.pc
	mov	ecx, 6
.l1	lodsb
	out	dx, al
	inc	dx
	loop	.l1
	mov	al, 0x08	; Activate -IEN (enable IRQ14)
	mov	dx, control
	out	dx, al
;pchr '*'
	call	ide_drdy_wait
	mov byte [done], 0	; clear IRQ flag
	DPOP	eax
	mov	dx, cbr.command
	out	dx, al		; send command byte

	mpop	eax, ecx, edx, esi
	ret


HDR:		;; ( start count addr -- [ior] )
		;; start = starting sector (LBA - linear addressing)
		;; count = number of sectors
		;; addr = address to load at
	mpush	eax, ebx, ecx, edx, esi, edi
	DPOP	edi		; addr
	DPOP	ecx		; count
	DPOP	eax		; start
	mov	esi,ecx

	cmp byte [valid], 0	; Make sure disk parameters are valid
	jz	.done

	;; Calculate sector count
	cmp	ecx, 256
	jb	.c1
	ja near .done		; too many sectors (To do: break it up into 256-sector reads)
	xor	cl, cl	; represent 256 as 0
.c1	mov	[regs.sc], cl

	call	lba2chs
	mov	[regs.lba], ebx

	DPUSH	0x20		; Read w/ retry
	call	ide_cmd
	mov	ecx, esi	; # of sectors * 256
	;; Read sector(s) from buffer
	mov	dx, cbr.data	; dx = io port
	cld
.read	call	ide_wait_irq
	push	ecx
	mov	ecx, 256
	rep	insw		; read data via PIO
	pop	ecx
	loop	.read

.done	mpop	eax, ebx, ecx, edx, esi, edi
	ret
eHDR


HDW:		;; ( start count addr -- [ior] )
		;; start = starting sector (LBA - linear addressing)
		;; count = number of sectors
		;; addr = address to load at
	mpush	eax, ebx, ecx, edx, esi, edi
	DPOP	esi		; addr
	DPOP	ecx		; count
	DPOP	eax		; start
	mov	edi,ecx

	cmp byte [valid], 0	; Make sure disk parameters are valid
	jz	.done

	;; Calculate sector count
	cmp	ecx, 256
	jb	.c1
	ja near .done		; too many sectors (To do: break it up into 256-sector reads)
	xor	cl, cl	; represent 256 as 0
.c1	mov	[regs.sc], cl

	call	lba2chs
	mov	[regs.lba], ebx

	DPUSH	0x30		; Write w/ retry
	call	ide_cmd
	call	ide_wait_drq

	mov	ecx, edi	; # of sectors * 256
	mov	edx, cbr.data
	cld
.write	push	ecx
	mov	ecx, 256
	rep	outsw		; write data via PIO
	pop	ecx
	cmp	edi, 1		; single sector? no IRQ to wait for
	je	.done
	call	ide_wait_irq
	loop	.write

.done	mpop	eax, ebx, ecx, edx, esi, edi
	ret
eHDW


lba2chs:
;;
;; Translate LBA ==> CHS
;;
;; Input:	EAX = lba sector
;; Output:	EBX = s:cy:h
;; Changes:	EAX, EBX, EDX
;;
	xor	edx, edx
	div dword [dsect]
	inc	dl
	mov	bh, dl
;	mov	[regs.sn], dl
;pnum edx			;dx=sect
	xor	edx, edx
	div dword [dhead]	;dx=head, ax=cyl
;pnum edx
;pnum eax
	or	dl, 0xA0
	mov	bl, dl
	rol	ebx, 16
	mov	bx, ax
	rol	ebx, 8
;	mov	[regs.dh], dl
;	mov	[regs.cy], ax
;phex ebx
;cr
	ret


ide_id:
;;
;; Identify drive & parse results
;;
	pusha
	DPUSH	0xEC		; Identify Drive command
	call	ide_cmd
	call	ide_wait_irq
;	or	eax, eax
;	jnz near .err

	mov	ecx, 256	; read data
	mov	edi, 0x4000
	mov	edx, cbr.data
	cld
	rep	insw

	sprint	"Model: "
	mov	esi, 0x4000 + 0x1b*2
	mov	ecx, 40
.l1	lodsb				;model
	pchr	eax
	loop	.l1
	dcr

	mov	edi, 0x4000
	xor	eax, eax
	sprint	"CHS="
	mov	ax, [edi+0x01*2]	;cyls
	mov	[dcyl], eax
	pnum	eax
	pchr	'/'
	mov	ax, [edi+0x03*2]	;heads
	mov	[dhead], eax
	pnum	eax
	pchr	'/'
	mov	ax, [edi+0x06*2]	;sectors
	mov	[dsect], eax
	pnum	eax
	pchr	','
	dspace

	mov	ax, [edi+0x15*2]	;buffer size
	pnum	eax
	sprint	"-sector buffer, pio>="

	mov	ax, [edi+0x33*2]	;Min. PIO
	pnum	eax
	sprint	", dma>="
	mov	ax, [edi+0x34*2]	;Min. DMA
	pnum	eax

	sprint	", controller type="
	mov	ax, [edi+0x14*2]
	pnum	eax
	dcr

	sprint	"Configuration bitmask: "
	mov	ax, [edi]
	mpush	[base]
	mov dword [base], 2
	DPUSH	eax, 16+0x80	;16 digits, 0-padded
	call	print_num_pad
	mpop	[base]
	dcr

	mov	al, 1
.done	mov byte [valid], al
	popa
	ret
.err	sprint	"Couldn't identify drive, hard disk access disabled.",10
	xor	al, al
	jmp short .done



;; MSDOS filesystem stuff - move out of here later..
;
; I want to load directories & fat entries only when they're needed...
; Use a sector cache (beware of 12-bit FAT entries crossing sectors!)

	SECTION .bss align=4

ptn
.lbn	resd 1	;partition LBN
.size	resd 1	;	size
.type	resb 1	;	type (fat12/fat16/linux/etc..)
.clus	resb 1	;sect/cluster
.fats	resw 1	;sect/FAT
.fat1	resd 1	;LBN of FAT1
.root	resd 1	;LBN of root dir
.roote	resd 1	;root dir size (in entries, 32 bytes each)
.roots	resd 1	;root dir size (in sectors)
.data	resd 1	;LBN of data area

	SECTION .text

SCRATCH	equ	0x40000

SECTOR	;; ( n -- addr , load sector via cache )
	ret
eSECTOR


INIT_FAT	;; run at boot time (or driver init, whatever)
	mpusha
	mov	esi, SCRATCH
	DPUSH	0, 1, esi		;get MBR
	call	HDR
	add	esi, 0x1be + 4	;1st ptn entry
	lodsd
	mov	[ptn.type], al		;Type
	lodsd
	mov	[ptn.lbn], eax		;LBN - start sector
	mov	ebx, eax		;save for later
	lodsd
	mov	[ptn.size], eax		;Size - in sectors

	mov	esi, SCRATCH
	DPUSH	ebx, 1, esi		;get BB (ptn. boot block)
	call	HDR
	movzx	eax, byte [esi+13]	;sect/cluster
	mov	[ptn.clus], al
	mov	ax, [esi+14]		;reserved sectors before FAT1
	add	ebx, eax
	mov	[ptn.fat1], ebx
	movzx	eax, byte [esi+16]	;# of fats
	movzx	ecx, word [esi+22]	;sect/FAT
	mov	[ptn.fats], cx
	mul	ecx
	add	ebx, eax		;calculate LBN of root dir
	mov	[ptn.root], ebx
	movzx	eax, word [esi+17]	;root dir entries
	mov	[ptn.roote], eax
	dec	eax
	shr	eax, 4			;calculate size of root dir (sectors)
	inc	eax
	mov	[ptn.roots], eax
	add	ebx, eax
	mov	[ptn.data], ebx
;DPUSH ptn	;return table address (temporarily)
	mpopa
	ret
eINIT_FAT


PARSE_FN	;; ( "file.ext" -- string , output in DOS 8+3 format )
		;; Output appears as in a directory entry - 11 chars, no '.'
		;; This routine allows 11-char filenames with no extension,
		;; which is what I prefer. Extensions are useless.
	mpusha
	CALL	PARSE_WORD
	DPOP	edx			;length
	DPOP	esi			;address
	add	edx, esi		;end of string
	mov	edi, [_HERE]
	DPUSH	edi
	mov	ecx, 8			;Filename
	call	.copy
	mov	cl, 3			;Extension
	call	.copy
	add	dword [_HERE], byte 11
	mpopa
	ret

.copy	cmp	esi, edx
	jae	.pad
	lodsb
	cmp	al, '.'
	je	.pad
				;todo - convert to UPPERCASE
	stosb
	loop	.copy
.pad	mov	al, ' '
	rep	stosb
	ret
ePARSE_FN


SEARCH_FN	;; ( string -- cluster | 0 )
		;; Search for file, return starting cluster or 0 if not found
	mpusha
	MDPUSH	[ptn.root]
	MDPUSH	[ptn.roots]
	DPUSH	SCRATCH
	call	HDR			;Load root directory
	mov	ebx, [ptn.roote]
	xor	ecx, ecx
	mov	edi, SCRATCH
.loop	mov	esi, S(0)
	mov	cl, 11
	rep	cmpsb
	je	.found
	add	cl, 21			;next entry
	add	edi, ecx
	dec	ebx
	jnz	.loop
	mov	S(0), ebx		;ebx=0
.done	mpopa
	ret
.found	mov	cx, [edi+0xF]
	mov	S(0), ecx
	jmp short .done
eSEARCH_FN


LOAD_FILE	;; ( cluster -- addr )
	mpusha
	MDPUSH	[ptn.fat1]
	movzx	eax, word [ptn.fats]
	DPUSH	eax
	DPUSH	SCRATCH
	call	HDR			;Load FAT #1
;DPUSH SCRATCH
;CALL MD
	mpush	S(0)
.next	pop	ebx			;cluster #
	mov	esi, ebx		;first cluster in sequence
	mov	eax, ebx		;last in sequence
	
.0	cmp byte [ptn.type], 1		;FAT12 or FAT16? (or other...?)
	jne	.fat16
	inc	eax

.fat12	cmp	bx, 0xFF0
	jae	.2
	mov	edi, ebx
	rcr	bx, 1			;set carry if odd
					;bx = 0x8000 + cluster/2
					;ebx+edi = 1.5*edi
	mov	bx, [ebx+edi+SCRATCH-0x8000]	;Get next cluster
	jnc	.even
	shr	ebx, 4		;only if odd
.even	and	bh, 0xF
	jmp short .common

.fat16	cmp	bx, 0xFFF0
	jae	.2
	inc	eax
	mov	bx, [ebx+SCRATCH]		;Get next cluster

.common	cmp	eax, ebx	;is next cluster contiguous?
	je	.0
.2	sub	eax, esi	;eax=count
	jz	.eof
	push	ebx		;save next cluster (EOF or discontiguous)

;...load the sectors
;  need to account for cluster size > 1 sector
	DPUSH	esi		;first sector
	DPUSH	eax		;count
;	DPUSH			address
;	call	HDR
;	jmp short .next
.eof
	mpopa
	ret
eLOAD_FILE
